'''
   File name: ReprtLib.py
   Author: Wayne Parham
   Date created: 04/01/2018
'''


from RulesLib import *


def ReadManifestFile( manifestfile ):
   data = []
   lines = []
   keycount = 0
   linecount = 0
   datatype = ""
   print( "==> Processing [" + manifestfile + "] File." )
   try:
      manifest = open( manifestfile, "r" )
      try:
         data = json.load(manifest)
         datatype = "JSON"
         print( "==> Manifest File [" + manifestfile + "] contains JSON data." )
      except ValueError:
         manifest.seek(0)
         for line in manifest:
            line = line.rstrip()
            if (len(line) > 0):
               if (line[0] != "0"):
                  if (line[0] != "#"):
                     lines.append( line )
                     keycount += 1
            linecount += 1
         manifest.close
         data = lines
         print( "==> Manifest File [" + manifestfile + "] contains (" + str(linecount) + ") lines, (" + str(keycount) + ") containing key/value fields." )
         if( cleanup_enabled ):
            try:
               os.remove( manifestfile )
            except Exception: 
               pass
   except IOError:
      print( "==> ERROR: [" + manifestfile + "] manifest file can't be opened for reading." )
   return( data, datatype )


def ProcessManifest( lines ): 
   catalog = []
   exhibits = []
   ruleset_steps = []
   ruleset_exhibits = []
   ruleset_summmary = []
   process_steps = 0
   process_exhibits = 1
   iteration = 0
   ruleset_status = ""
   ruleset_start = ""
   ruleset_end = ""
   last_line = len(lines)
   try:
      while( iteration < 3 ):
         linecount = 0
         for line in lines:
            location = []
            position = 0
            linecount += 1
            end = len( line )
            while( position < end ):
               position = ( line.find( ':', position, end ) )
               if( position != -1 ):
                  location.append( position )
                  position += 1
               else:
                  position = end
            count = len( location )
            if( count > 0 ):
               separator = location[0]
               key = line[0:separator].strip()
               value = line[(separator+1):end].strip()
            if( iteration == process_steps ):
               if( key.lower() == "steps" ):
                  step_data = ParseSteps( value )
                  if( len(step_data) == 3 ):
                     ruleset_steps = step_data[0]
                     ruleset_exhibits = step_data[1]
                     ruleset_summary = step_data[2]
                     ruleset_summary_count = len( ruleset_summary )
                     ruleset_exhibit_count = len( ruleset_exhibits )
                     if( ruleset_summary_count > 0 ):
                        i = 0
                        while( i < ruleset_summary_count ):
                           summary_key = ruleset_summary[i]
                           summary_value = ruleset_summary[i+1]
                           if( summary_key.lower() == "status" ):
                              ruleset_status = summary_value
                           if( summary_key.lower() == "start" ):
                              ruleset_start = summary_value
                           if( summary_key.lower() == "end" ):
                              ruleset_end = summary_value
                           i += 2
                  else:
                     print( "==> ERROR: Invalid Step Object." )
                     catalog = [-1]
            elif( iteration == process_exhibits ):
               if( key.lower() == "evidences" ):
                  exhibits = ParseExhibits( value )
                  ruleset_exhibit_count = len( ruleset_exhibits )
                  if( ruleset_exhibit_count > 0 ):
                     i = 0
                     while( i < ruleset_exhibit_count ):
                        exhibits.append( ruleset_exhibits[i] )
                        i += 1
            else:
               if( (key.lower() != "steps") and (key.lower() != "evidences") ):
                  if( key.lower() == "status" ):
                     if( ((len(value) == 0) or (value.lower() == "todo") or (value.lower() == "pass")) and (len(ruleset_status) > 0) ):
                        value = ruleset_status
                  if( (key.lower() == "start") and (len(value) == 0) ):
                        value = ruleset_start
                  if( (key.lower() == "finish") and (len(value) == 0) ):
                        value = ruleset_end
                  catalog.append( key )
                  catalog.append( value )
         iteration += 1
      if( len(exhibits) > 0 ):
         catalog.append( "evidences" )
         catalog.append( exhibits )
      if( len(ruleset_steps) > 0 ):
         catalog.append( "steps" )
         catalog.append( ruleset_steps )
   except:
      print( "==> ERROR: Bad Manifest File.  Malformed or format not recognized." )
   return( catalog )


def ProcessJSON( data ):
   cleanup_list = []
   info = []
   message = ""
   info_end = ""
   info_start = ""
   info_status = ""
   overall_end = ""
   overall_start = ""
   overall_status = ""
   needs_info_start = True
   needs_info_finish = True
   for key in data:
      value = data[key]
      if( key.lower() == "info" ):
         for infokey in value:
            infovalue = value[infokey]
            info.append( infokey )
            info.append( infovalue )
      elif( key.lower() == "tests" ):
         i = 0
         test_number = 0
         for tests in value:
            test_number += 1
            info_count = 0
            test_count = 0
            info_end = ""
            info_start = ""
            info_status = ""
            info_summary = ""
            overall_status = ""
            test = []
            steps = []
            summary = []
            evidence = []
            ruleslist = []
            ruleset_steps = []
            ruleset_summary = []
            ruleslist_steps = []
            ruleslist_summary = []
            ruleslist_steps_array = []
            step_ruleset_steps = []
            step_ruleset_summary = []
            step_ruleslist_steps = []
            step_ruleslist_summary = []
            step_ruleslist_steps_array = []
            needs_test_start = True
            needs_test_finish = True
            needs_test_status = True
            data_present = True
            filename = ""
            filedata = ""
            filetype = ""
            comment = ""
            status = ""
            try:
               for testkey in data["tests"][i]:
                  testvalue = data["tests"][i][testkey]
                  if( testkey.lower() == "evidences" ):
                     evidence_present = True
                     e = 0
                     while( evidence_present ):
                        try:
                           for evidencekey in data["tests"][i]["evidences"][e]:
                              evidencevalue = data["tests"][i]["evidences"][e][evidencekey]
                              if( evidencekey.lower() == "filename" ):
                                 filename = evidencevalue
                              if( evidencekey.lower() == "data" ):
                                 if( evidencevalue ):
                                    filedata = evidencevalue
                                 else:
                                    filedata = ":"
                              if( filename and filedata ):
                                 pathname = datadir+"/"+filename
                                 if( filedata == ":" ):
                                    filedata = EncodeFile( pathname )
                                 else:
                                    WriteEncodedFile( pathname, filedata )
                                 evidence.append( "filename" )
                                 evidence.append( filename )
                                 evidence.append( "data" )
                                 evidence.append( filedata )
                                 cleanup_list.append( pathname )
                                 filename = ""
                                 filedata = ""
                           e += 1
                        except IndexError:
                           evidence_present = False
                  elif( testkey.lower() == "steps" ):
                     results_present = False
                     rules_present = False
                     steps_present = True
                     actionable = False
                     s = 0
                     while( steps_present ):
                        try:
                           for stepskey in data["tests"][i]["steps"][s]:
                              stepsvalue = data["tests"][i]["steps"][s][stepskey]
                              if( stepskey.lower() == "evidences" ):
                                 step_rulesets_present = True
                                 z = 0
                                 while( step_rulesets_present ):
                                    actionable = False
                                    step_ruleslist_steps = []
                                    step_ruleslist_summary = []
                                    try:
                                       for step_ruleslistkey in data["tests"][i]["steps"][s]["evidences"][z]:
                                          step_ruleslistvalue = data["tests"][i]["steps"][s]["evidences"][z][step_ruleslistkey]
                                          if( step_ruleslistkey.lower() == "filename" ):
                                             filename = step_ruleslistvalue
                                          if( step_ruleslistkey.lower() == "contenttype" ):
                                             filetype = step_ruleslistvalue
                                          if( step_ruleslistkey.lower() == "data" ):
                                             if( step_ruleslistvalue ):
                                                filedata = step_ruleslistvalue
                                             else:
                                                filedata = ":"
                                          if( filename and filetype and filedata ):
                                             actionable = True
                                             pathname = datadir+"/"+filename
                                             if( filedata == ":" ):
                                                filedata = EncodeFile( pathname )
                                                cleanup_list.append( pathname )
                                             else:
                                                WriteEncodedFile( pathname, filedata )
                                                cleanup_list.append( pathname )
                                             if( filename.lower().endswith(rules_identifier) ):
                                                step_ruleslist_steps.insert(0, filedata )
                                                step_ruleslist_steps.insert(0, "data")
                                                step_ruleslist_steps.insert(0, "application/csv" )
                                                step_ruleslist_steps.insert(0, "contentType")
                                                step_ruleslist_steps.insert(0, filename )
                                                step_ruleslist_steps.insert(0,"filename")
                                                results = RunRules( pathname )
                                                if( results != [-1] ):
                                                   print( "==> Ran [" + pathname + "] through Rules Engine, now building Report." )
                                                   step_ruleslist_summary, step_ruleslist_steps = ParseRulesResults( results )
                                                   x = 0
                                                   info_count = len( info )
                                                   test_count = len( test )
                                                   summary_count = len( step_ruleslist_summary )
                                                   if( summary_count > 0 ):
                                                      while( x < summary_count ):
                                                         if( step_ruleslist_summary[x].lower() == "start" ):
                                                            if( not info_start ):
                                                               info_start = step_ruleslist_summary[x+1].replace(" ", "T", 1) + "+0000"
                                                               if( not overall_start):
                                                                  overall_start = info_start
                                                            if( step_ruleslist_summary[x].lower() == "end" ):
                                                               info_end = step_ruleslist_summary[x+1].replace(" ", "T", 1) + "+0000"
                                                               overall_end = info_end
                                                            if( step_ruleslist_summary[x].lower() == "status" ):
                                                               info_status = step_ruleslist_summary[x+1]
                                                            if( (not overall_status) or (overall_status == "PASS") ):
                                                               overall_status = info_status
                                                         x += 1
                                                   step_ruleslist_steps_array.append( step_ruleslist_steps )
                                                   rulesfile = filename.rsplit(".csv", 1)[0]
                                                   logfile = rulesfile+".log"
                                                   logpath = logdir+"/"+logfile
                                                   logdata = EncodeFile( logpath )
                                                   cleanup_list.append( logpath )
                                                   if( logdata != -1 ):
                                                      evidence.append("filename")
                                                      evidence.append( logfile )
                                                      evidence.append("contentType")
                                                      evidence.append("text/plain")
                                                      evidence.append("data")
                                                      evidence.append( logdata )
                                             else:
                                                step_ruleslist_entry = []
                                                step_ruleslist_entry.append("filename")
                                                step_ruleslist_entry.append( filename )
                                                step_ruleslist_entry.append("contentType")
                                                step_ruleslist_entry.append( filetype )
                                                step_ruleslist_entry.append("data")
                                                step_ruleslist_entry.append( filedata )
                                                step_ruleslist.append( step_ruleslist_entry )
                                             actionable = False
                                             filetype = ""
                                             filename = ""
                                             filedata = ""
                                       z += 1
                                    except IndexError:
                                       step_rulesets_present = False
                                       actionable = True
                              if( stepskey.lower() == "filename" ):
                                 filename = stepsvalue
                              if( stepskey.lower() == "data" ):
                                 if( stepsvalue ):
                                    filedata = stepsvalue
                                 else:
                                    filedata = ":"
                              if( stepskey.lower() == "contenttype" ):
                                 filetype = stepsvalue
                                 if( actionable ):
                                    steps.append("contentType")
                                    steps.append( filetype )
                              if( stepskey.lower() == "status" ):
                                 status = stepsvalue
                                 results_present = True
                                 if( actionable ):
                                    steps.append("status")
                                    steps.append( status )
                              if( stepskey.lower() == "comment" ):
                                 comment = stepsvalue
                                 results_present = True
                                 if( actionable ):
                                    steps.append("comment")
                                    steps.append( comment )
                              if( actionable and filename and filedata ):
                                 rules_present = True
                                 steps.append("filename")
                                 steps.append( filename )
                                 steps.append("data")
                                 pathname = datadir+"/"+filename
                                 if( filedata == ":" ):
                                    steps.append( EncodeFile( pathname ) )
                                    cleanup_list.append( pathname )
                                 else:
                                    steps.append( filedata )
                                    WriteEncodedFile( pathname, filedata )
                                    cleanup_list.append( pathname )
                              if( actionable and rules_present and not results_present and filename.lower().endswith(rules_identifier) ):
                                 results = RunRules( pathname )
                                 if( results != [-1] ):
                                    print( "==> Ran [" + pathname + "] through Rules Engine, now building Report." )
                                    ruleset_summary, ruleset_steps = ParseRulesResults( results )
                                    rulesfile = filename.rsplit(".csv", 1)[0]
                                    logfile = rulesfile+".log"
                                    logpath = logdir+"/"+logfile
                                    logdata = EncodeFile( logpath )
                                    cleanup_list.append( logpath )
                                    if( logdata != -1 ):
                                       evidence.append("filename")
                                       evidence.append( logfile )
                                       evidence.append("contentType")
                                       evidence.append("text/plain")
                                       evidence.append("data")
                                       evidence.append( logdata )
                                    filename = ""
                                    filedata = ""
                                    filetype = ""
                           s += 1
                           filename = ""
                           filedata = ""
                           filetype = ""
                           comment = ""
                           status = ""
                        except IndexError:
                           if( not actionable ):
                              actionable = True
                              s = 0
                           else:
                              steps_present = False
                              summary_count = len( ruleset_summary )
                              info_count = len( info )
                              test_count = len( test )
                              ruleset_stepcount = len( ruleset_steps )
                              x = 0
                              while( x < ruleset_stepcount ):
                                 steps.append( ruleset_steps[x] )
                                 x += 1
                              x = 0
                              if( summary_count > 0 ):
                                 while( x < summary_count ):
                                    if( ruleset_summary[x].lower() == "start" ):
                                       if( not info_start ):
                                          info_start = ruleset_summary[x+1].replace(" ", "T", 1) + "+0000"
                                          if( not overall_start ):
                                             overall_start = info_start
                                    if( ruleset_summary[x].lower() == "end" ):
                                       info_end = ruleset_summary[x+1].replace(" ", "T", 1) + "+0000"
                                       overall_end = info_end
                                    if( ruleset_summary[x].lower() == "status" ):
                                       info_status = ruleset_summary[x+1]
                                       if( (not overall_status) or (overall_status == "PASS") ):
                                          overall_status = info_status
                                    x += 1
                  elif( testkey.lower() == "rules" ):
                     rulesets_present = True
                     t = 0
                     while( rulesets_present ):
                        actionable = True
                        ruleslist_steps = []
                        ruleslist_summary = []
                        try:
                           for ruleslistkey in data["tests"][i]["rules"][t]:
                              ruleslistvalue = data["tests"][i]["rules"][t][ruleslistkey]
                              if( ruleslistkey.lower() == "filename" ):
                                 filename = ruleslistvalue
                              if( ruleslistkey.lower() == "contenttype" ):
                                 filetype = ruleslistvalue
                              if( ruleslistkey.lower() == "data" ):
                                 if( ruleslistvalue ):
                                    filedata = ruleslistvalue
                                 else:
                                    filedata = ":"
                              if( filename and filetype and filedata ):
                                 actionable = True
                                 pathname = datadir+"/"+filename
                                 if( filedata == ":" ):
                                    filedata = EncodeFile( pathname )
                                    cleanup_list.append( pathname )
                                 else:
                                    WriteEncodedFile( pathname, filedata )
                                    cleanup_list.append( pathname )
                                 if( filename.lower().endswith(rules_identifier) ):
                                    ruleslist_steps.insert(0, filedata )
                                    ruleslist_steps.insert(0, "data")
                                    ruleslist_steps.insert(0, "application/csv" )
                                    ruleslist_steps.insert(0, "contentType")
                                    ruleslist_steps.insert(0, filename )
                                    ruleslist_steps.insert(0,"filename")
                                    results = RunRules( pathname )
                                    if( results != [-1] ):
                                       print( "==> Ran [" + pathname + "] through Rules Engine, now building Report." )
                                       ruleslist_summary, ruleslist_steps = ParseRulesResults( results )
                                       x = 0
                                       info_count = len( info )
                                       test_count = len( test )
                                       summary_count = len( ruleslist_summary )
                                       if( summary_count > 0 ):
                                          while( x < summary_count ):
                                             if( ruleslist_summary[x].lower() == "start" ):
                                                if( not info_start ):
                                                   info_start = ruleslist_summary[x+1].replace(" ", "T", 1) + "+0000"
                                                   if( not overall_start ):
                                                      overall_start = info_start
                                             if( ruleslist_summary[x].lower() == "end" ):
                                                info_end = ruleslist_summary[x+1].replace(" ", "T", 1) + "+0000"
                                                overall_end = info_end
                                             if( ruleslist_summary[x].lower() == "status" ):
                                                info_status = ruleslist_summary[x+1]
                                                if( (not overall_status) or (overall_status == "PASS") ):
                                                   overall_status = info_status
                                             x += 1
                                       ruleslist_steps_array.append( ruleslist_steps )
                                       rulesfile = filename.rsplit(".csv", 1)[0]
                                       logfile = rulesfile+".log"
                                       logpath = logdir+"/"+logfile
                                       logdata = EncodeFile( logpath )
                                       cleanup_list.append( logpath )
                                       if( logdata != -1 ):
                                          evidence.append("filename")
                                          evidence.append( logfile )
                                          evidence.append("contentType")
                                          evidence.append("text/plain")
                                          evidence.append("data")
                                          evidence.append( logdata )
                                 else:
                                    ruleslist_entry = []
                                    ruleslist_entry.append("filename")
                                    ruleslist_entry.append( filename )
                                    ruleslist_entry.append("contentType")
                                    ruleslist_entry.append( filetype )
                                    ruleslist_entry.append("data")
                                    ruleslist_entry.append( filedata )
                                    ruleslist.append( ruleslist_entry )
                                 actionable = False
                                 filetype = ""
                                 filename = ""
                                 filedata = ""
                           t += 1
                        except IndexError:
                           rulesets_present = False
                  else:
                     test.append( testkey )
                     test.append( testvalue )
               i += 1
            except IndexError:
               data_present = False
            steps_size = len(steps)
            ruleslist_steps_count = len(ruleslist_steps_array)
            step_ruleslist_steps_count = len(step_ruleslist_steps_array)
            combined_steps = []
            if( step_ruleslist_steps_count > 0 ):
               x = 0
               while( x < step_ruleslist_steps_count ):
                  combined_steps.append( step_ruleslist_steps_array[x] )
                  x += 1
            if( steps_size > 0 ):
               x = 0
               while( x < steps_size ):
                  combined_steps.append( steps[x] )
                  x += 1
            if( ruleslist_steps_count > 0 ):
               x = 0
               while( x < ruleslist_steps_count ):
                  combined_steps.append( ruleslist_steps_array[x] )
                  x += 1
            steps = combined_steps
            x = 0
            while( x < info_count ):
               if( info[x].lower() == "startdate" ):
                  info[x+1] = overall_start
                  needs_info_start = False
               if( info[x].lower() == "finishdate" ):
                  info[x+1] = overall_end
                  needs_info_finish = False
               x += 1
            if( needs_info_start ):
               info.append("startDate")
               info.append( overall_start )
               needs_info_start = False
            if( needs_info_finish ):
               info.append("finishDate")
               info.append( overall_end )
               needs_info_finish = False
            x = 0
            while( x < test_count ):
               if( test[x].lower() == "start" ):
                  test[x+1] = info_start
                  needs_test_start = False
               if( test[x].lower() == "finish" ):
                  test[x+1] = info_end
                  needs_test_finish = False
               if( test[x].lower() == "status" ):
                  test[x+1] = overall_status
                  needs_test_status = False
               x += 1
            if( needs_test_start ):
               test.append("start")
               test.append( info_start )
               needs_test_start = False
            if( needs_test_finish ):
               test.append("finish")
               test.append( info_end )
               needs_test_finish = False
            if( needs_test_status ):
               test.append("status")
               test.append( overall_status )
               needs_test_status = False
            json_test = FormatTest( test )
            json_steps = FormatSteps( steps )
            if( attachments_enabled ):
               json_evidence = FormatEvidence( evidence )
            json_ruleslist = FormatRules( ruleslist )
            stepblock = False
            exibblock = False
            listblock = False
            if( len(steps) > 0 ):
               stepblock = True
            if( len(evidence) > 0 ):
               exibblock = True
            if( len(ruleslist) > 0 ):
               listblock = True
            if( test_number == 1 ):
              message += "    " + quote + "tests" + quote + ": [{\n"
            else:
              message += "\n    },\n    {\n"
            message += json_test
            if( attachments_enabled ):
               if( exibblock ):
                  message += ",\n"
                  message += json_evidence
            if( stepblock ):
               message += ",\n"
               message += json_steps
            if( listblock ):
               message += ",\n"
               message += json_ruleslist
   message += "\n    }]\n  }\n"
   json_info = FormatInfo( info )
   message = "  {\n" + json_info + ",\n" + message
   if( cleanup_enabled ):
      for pathname in cleanup_list:
         try:
            os.remove( pathname )
         except Exception: 
            pass
   return( message )


def FormatInfo( info ):
   count = len( info )
   json_info = "    " + quote + "info" + quote + ": {\n"
   i = 0
   while( i < count ):
      if( i > 0 ):
         json_info += ",\n" 
      json_info += "      " + quote + info[i] + quote + ": "
      json_info += quote + info[i+1] + quote
      i += 2
   json_info += "\n    }"
   return( json_info )


def FormatTest( test ):
   count = len( test )
   json_test = ""
   i = 0
   while( i < count ):
      if( i > 0 ):
         json_test += ",\n" 
      json_test += "      " + quote + test[i] + quote + ": "
      if( type(test[i+1]) == list ):
         print( "==> WARNING: Unidentified value '"+test[i]+"' in metadata. Check "+manifest_file+" file." )
      json_test += quote + str(test[i+1]) + quote
      i += 2
   return( json_test )


def FormatEvidence( evidence ):
   contents = []
   json_evidence = "      " + quote + "evidences" + quote + ": [{\n"
   exibcount = len( evidence )
   j = 0
   while( j < exibcount ):
      if( evidence[j] in contents ):
         json_evidence += "\n        },\n        {\n"
         contents = []
      elif( j > 0 ):
         json_evidence += ",\n"
      contents.append( evidence[j] )
      json_evidence += "        " + quote + evidence[j] + quote + ": "
      json_evidence += quote + evidence[j+1] + quote
      j += 2
   json_evidence += "\n        }\n      ]"
   return( json_evidence )


def FormatSteps( steps ):
   contents = []
   results_found = 0
   json_steps = "      " + quote + "steps" + quote + ": [{\n"
   if ( IsListofLists(steps) ):
      arraycount = len( steps )
      last_step = ""
      i = 0
      while( i < arraycount ):
         if( i > 0 ):
            json_steps += ",\n        {\n"
         step = steps[i]
         stepcount = len( step )
         j = 0
         while( j < stepcount ):
            if( ((step[j].lower() == "comment") or (step[j]).lower() == "status") and not results_found ):
               if( j > 0 ):
                  json_steps += "\n        },\n        {\n"
               contents = []
               results_found = 1
            elif( (step[j].lower() != "comment") and (step[j].lower() != "status") and results_found ):
               if( j > 0 ):
                  json_steps += "\n        },\n        {\n"
               contents = []
               results_found = 0
            elif( (last_step != "filename") and (last_step != "data") and (last_step != "contenttype") ):
               contents = []
               if( (step[j].lower() == "comment") or (step[j].lower() == "status") ):
                  json_steps += "\n        },\n        {\n"
            elif( j > 0  ):
               json_steps += ",\n"
            contents.append( step[j] )
            last_step = step[j].lower()
            step_text = str(step[j+1]).replace( '\\', '\\\\' ).replace( '"', r'\"' ).replace( '\0', '' )
            if( (step[j].lower() != "filename") and (step[j].lower() != "data") and (step[j].lower() != "contenttype") ):
               json_steps += "        " + quote + step[j] + quote + ": "
               json_steps += quote + step_text + quote
            j += 2
         json_steps += "\n        }"
         i += 1
      json_steps += "\n      ]"
   else:
      stepcount = len( steps )
      last_step = ""
      j = 0
      while( j < stepcount ):
         if( ((steps[j].lower() == "comment") or (steps[j].lower() == "status")) and not results_found ):
            json_steps += "\n        },\n        {\n"
            contents = []
            results_found = 1
         elif( ((steps[j].lower() != "comment") and (steps[j].lower() != "status") and results_found)  ):
            json_steps += "\n        },\n        {\n"
            contents = []
            results_found = 0
         elif( (steps[j] in contents) and ( (last_step != "filename") and (last_step != "data") and (last_step != "contenttype") ) ):
            json_steps += "\n        },\n        {\n"
            contents = []
         elif( j > 0 ):
            if( (last_step != "filename") and (last_step != "data") and (last_step != "contenttype") ):
               json_steps += ",\n"
         contents.append( steps[j] )
         last_step = steps[j].lower()
         try:
            step_text = str(steps[j+1]).replace( '\\', '\\\\' ).replace( '"', r'\"' ).replace( '\0', '' )
         except IndexError:
            print( "==> ERROR: Malformed Steps container.  Please check manifest file." )
         if( (steps[j].lower() != "filename") and (steps[j].lower() != "data") and (steps[j].lower() != "contenttype") ):
           json_steps += "        " + quote + steps[j] + quote + ": "
           json_steps += quote + step_text + quote
         j += 2
      json_steps += "\n        }\n      ]"
   return( json_steps )


def FormatRules( ruleslist ):
   json_ruleslist = "      " + quote + "rules" + quote + ": [{\n"
   if ( IsListofLists(ruleslist) ):
      arraycount = len( ruleslist )
      i = 0
      while( i < arraycount ):
         if( i > 0 ):
            json_ruleslist += ",\n        {\n"
         test = ruleslist[i]
         testcount = len( test )
         j = 0
         while( j < testcount ):
            if( j > 0 ):
               json_ruleslist += ",\n"
            json_ruleslist += "        " + quote + test[j] + quote + ": "
            json_ruleslist += quote + test[j+1] + quote
            j += 2
         json_ruleslist += "\n        }"
         i += 1
      json_ruleslist += "\n      ]"
   else:
      testcount = len( ruleslist )
      contents = []
      j = 0
      while( j < testcount ):
         if( ruleslist[j] in contents ):
            json_ruleslist += "\n        },\n        {\n"
            contents = []
         elif( j > 0 ):
            json_ruleslist += ",\n"
         contents.append( ruleslist[j] )
         json_ruleslist += "        " + quote + ruleslist[j] + quote + ": "
         json_ruleslist += quote + ruleslist[j+1] + quote
         j += 2
      json_ruleslist += "\n        }\n      ]"
   return( json_ruleslist )


def ParseRulesResults( results ):
   if( results == [-1] ):
      summary = [-1]
      steps = [-1]
   else:
      j = 0
      steps = []
      summary = []
      result_count = len( results )
      while( j < result_count ):
         results_known = 0
         result_entry = results[j]
         if( result_entry[:6] == "start:" ):
            time_start = result_entry[-19:]
            summary.append( "start" ) 
            summary.append( time_start ) 
         elif( result_entry[:4] == "end:" ):
            status = result_entry[5:9]
            time_end = result_entry[-19:]
            summary.append( "end" ) 
            summary.append( time_end ) 
            summary.append( "status" ) 
            summary.append( status ) 
         elif( result_entry[:8] == "command:" ):
            command_entry = result_entry[9:]
            command_entry_value = command_entry.split(",,")
            sequence = command_entry_value[0]
            command = command_entry_value[1]
            status = "EXECUTING"
            comment = sequence + ". " + command
            results_known = 1
         elif( result_entry[:5] == "test:" ):
            test_entry = result_entry[6:]
            test_entry_value = test_entry.split(",,")
            sequence = test_entry_value[0]
            test = test_entry_value[1]
            status = ""
            comment = sequence + ". " + test
         elif( result_entry[:7] == "result:" ):
            result_entry = result_entry[8:]
            test_entry_value = result_entry.split(",,")
            sequence = test_entry_value[0]
            status = test_entry_value[1].strip()
            comment = comment + " at " + test_entry_value[2].strip()
            results_known = 1
         if( results_known ):
            steps.append("status")
            steps.append( status )
            steps.append("comment")
            steps.append( comment )
         j += 1
   return( summary, steps )


def ParseSteps( steps ):
   encoded_exhibits = []
   encoded_steps = []
   step_summary = []
   filenames = []
   keyvalue = []
   types = []
   data = []
   stat_present = 0
   cmnt_present = 0
   data_present = 0
   file_present = 0
   type_present = 0
   stat_position = 0
   cmnt_position = 0
   data_position = 0
   file_position = 0
   type_position = 0
   field = steps.split(",")
   fieldcount = len( field )
   if( fieldcount % 5 == 0 ):
      i = 0
      while( i < fieldcount ):
         pair = field[i]
         pair.strip()
         keyvalue = pair.split( ":" )
         if( len(keyvalue) != 2 ):
            junkdata = str( keyvalue )
            print( "==> ERROR: Steps syntax error.  Expected key:value pair." )
            print( "    " + junkdata )
            encoded_steps = [-1]
            break
         else:
            key = str(keyvalue[0]).strip()
            value = str(keyvalue[1]).strip()
            encoded_steps.append( key )
            encoded_steps.append( value )
         i += 1
      if( encoded_steps[0] != -1 ): 
         i = 0
         while( i < (fieldcount*2) ):
            key = str(encoded_steps[i])
            value = str(encoded_steps[i+1])
            if( key.lower() == "status" ):
               stat_position = i+1
               stat_present = len(value)
            if( key.lower() == "comment" ):
               cmnt_position = i+1
               cmnt_present = len(value)
            if( key.lower() == "data" ):
               data_position = i+1
               data_present = len(value)
            if( key.lower() == "filename" ):
               file_position = i+1
               file_present = len(value)
            if( key.lower() == "contenttype" ):
               type_position = i+1
               type_present = len(value)
            full_set_on_zero = ((i/2)+1) % 5
            if( full_set_on_zero == 0 ):
               if( data_present and not file_present ):
                  print( "==> ERROR: Steps error.  Data without filename." )
                  print( "    [" + steps + "]" )
                  encoded_steps = [-1]
               if( data_present and file_present):
                  filename = encoded_steps[file_position]
                  data = encoded_steps[data_position]
                  filepath = datadir+"/"+filename
                  WriteEncodedFile( filepath, data )
               if( file_present and not data_present ):
                  filename = encoded_steps[file_position]
                  filepath = datadir+"/"+filename
                  b64data = EncodeFile( filepath )
                  if( b64data != -1 ):
                     encoded_steps[data_position] = b64data
                  else:
                     encoded_steps = [-1]
               if( file_present and not type_present ):
                  encoded_steps[type_position] = DetermineFileType( encoded_steps[file_position] ) 
               if( file_present ):
                  filename = encoded_steps[file_position]
                  status   = encoded_steps[stat_position]
                  b64data  = encoded_steps[data_position]
                  if( filename.endswith(rules_identifier) and (status.lower() == "todo" or len(status) == 0) ):
                     contentType = "application/csv"
                     if( len(b64data) == 0 ):
                        filepath = datadir+"/"+filename
                        b64data = EncodeFile( filepath )
                        if( b64data != -1 ):
                           encoded_exhibits.append("data")
                           encoded_exhibits.append( b64data )
                           encoded_exhibits.append("filename")
                           encoded_exhibits.append( filename  )
                           encoded_exhibits.append("contentType")
                           encoded_exhibits.append( contentType )
                        else:
                           encoded_exhibits = [-1]
                     else:
                        encoded_exhibits.append("data")
                        encoded_exhibits.append( b64data )
                        encoded_exhibits.append("filename")
                        encoded_exhibits.append( filename  )
                        encoded_exhibits.append("contentType")
                        encoded_exhibits.append( contentType )
                     pathname = datadir+"/"+filename
                     results = RunRules( pathname )
                     if( results == -1 ):
                        encoded_steps = [-1]
                     else:
                        print( "==> Ran [" + pathname + "] through Rules Engine, now building Report." )
                        result_count = len( results )
                        j = 0
                        while( j < result_count ):
                           data = ""
                           filename = ""
                           contentType = ""
                           ready_to_enter = False
                           result_entry = results[j]
                           if( result_entry[:6] == "start:" ):
                              status = ""
                              time_start = result_entry[-19:]
                              comment = "Test start at " + time_start
                              step_summary.append( "start" ) 
                              step_summary.append( time_start ) 
                              ready_to_enter = True
                           elif( result_entry[:4] == "end:" ):
                              status = result_entry[5:9]
                              time_end = result_entry[-19:]
                              comment = "Test complete at " + time_end
                              step_summary.append( "end" ) 
                              step_summary.append( time_end ) 
                              step_summary.append( "status" ) 
                              step_summary.append( status ) 
                              ready_to_enter = True
                           elif( result_entry[:8] == "command:" ):
                              command_entry = result_entry[9:]
                              command_entry_value = command_entry.split(",,")
                              sequence = command_entry_value[0]
                              command = command_entry_value[1]
                              status = "EXECUTING"
                              comment = sequence + ". " + command
                              ready_to_enter = True
                           elif( result_entry[:5] == "test:" ):
                              test_entry = result_entry[6:]
                              test_entry_value = test_entry.split(",,")
                              sequence = test_entry_value[0]
                              test = test_entry_value[1]
                              status = ""
                              comment = sequence + ". " + test
                           elif( result_entry[:7] == "result:" ):
                              result_entry = result_entry[8:]
                              test_entry_value = result_entry.split(",,")
                              sequence = test_entry_value[0]
                              status = test_entry_value[1].strip()
                              comment = comment + " at " + test_entry_value[2].strip()
                              ready_to_enter = True
                           if( ready_to_enter ):
                              encoded_steps.append("status")
                              encoded_steps.append( status )
                              encoded_steps.append("comment")
                              encoded_steps.append( comment )
                              encoded_steps.append("data")
                              encoded_steps.append( "" )
                              encoded_steps.append("filename")
                              encoded_steps.append( "" )
                              encoded_steps.append("contentType")
                              encoded_steps.append( "" )
                           if( result_entry[:8] == "exhibit:" ):
                              filename = result_entry[9:].strip()
                              contentType = DetermineFileType( filename )
                              filepath = datadir+"/"+filename
                              b64data = EncodeFile( filepath )
                              if( b64data != -1 ):
                                 encoded_exhibits.append("data")
                                 encoded_exhibits.append( b64data )
                                 encoded_exhibits.append("filename")
                                 encoded_exhibits.append( filename  )
                                 encoded_exhibits.append("contentType")
                                 encoded_exhibits.append( contentType )
                              else:
                                 encoded_exhibits = [-1]
                           j += 1
               stat_present = 0
               cmnt_present = 0
               data_present = 0
               file_present = 0
               type_present = 0
            i += 2
   else:
      print( "==> ERROR: Steps syntax error.  Expected status, comment, data, filename and contentType." )
      print( "    [" + steps + "]" )
      encoded_steps = [-1]
   return( encoded_steps, encoded_exhibits, step_summary )


def ParseExhibits( exhibit ):
   encoded_exhibit = []
   filenames = []
   keyvalue = []
   types = []
   data = []
   data_present = 0
   file_present = 0
   type_present = 0
   data_position = 0
   file_position = 0
   type_position = 0
   field = exhibit.split(",")
   fieldcount = len( field )
   if( fieldcount % 3 == 0 ):
      i = 0
      while( i < fieldcount ):
         pair = field[i]
         pair.strip()
         keyvalue = pair.split( ":" )
         if( len(keyvalue) != 2 ):
            junkdata = str( keyvalue )
            print( "==> ERROR: Exhibit syntax error.  Expected key:value pair." )
            print( "    " + junkdata )
            encoded_exhibit = [-1]
            break
         else:
            key = str(keyvalue[0]).strip()
            value = str(keyvalue[1]).strip()
            encoded_exhibit.append( key )
            encoded_exhibit.append( value )
         i += 1
      if( encoded_exhibit[0] != -1 ): 
         i = 0
         while( i < (fieldcount*2) ):
            key = str(encoded_exhibit[i])
            value = str(encoded_exhibit[i+1])
            if( key.lower() == "data" ):
               data_position = i+1
               data_present = len(value) 
            if( key.lower() == "filename" ):
               file_position = i+1
               file_present = len(value) 
            if( key.lower() == "contenttype" ):
               type_position = i+1
               type_present = len(value) 
            full_set_on_zero = ((i/2)+1) % 3
            if( full_set_on_zero == 0 ):
               if( data_present and not file_present ):
                  print( "==> ERROR: Exhibit error.  Data without filename." )
                  print( "    [" + exhibit + "]" )
                  encoded_exhibit = [-1]
               if( data_present and file_present):
                  filename = encoded_exhibit[file_position]
                  data = encoded_exhibit[data_position]
                  filepath = datadir+"/"+filename
                  WriteEncodedFile( filepath, data )
               if( file_present and not data_present ):
                  filename = encoded_exhibit[file_position]
                  filepath = datadir+"/"+filename
                  b64data = EncodeFile( filename )
                  if( b64data != -1 ):
                     encoded_exhibit[data_position] = b64data
                  else:
                     encoded_exhibit = [-1]
               if( file_present and not type_present ):
                  encoded_exhibit[type_position] = DetermineFileType( encoded_exhibit[file_position] )
               data_present = 0
               file_present = 0
               type_present = 0
            i += 2
   else:
      print( "==> ERROR: Exhibit syntax error.  Expected data, filename and contentType." )
      print( "    [" + exhibit + "]" )
      encoded_exhibit = [-1]
   return( encoded_exhibit )


def CreateMessage( catalog ):
   i = 0
   j = 0
   retval = 0
   number_of_detail_entries = 0
   number_of_report_entries = 0
   number_of_catalog_entries = len( catalog )
   data = "{    " + quote + "info" + quote + ": {\n"
   while( i < number_of_catalog_entries ):
      key = catalog[i]
      value = catalog[i+1]
      i += 2
      if( (key.lower() == "summary") or
          (key.lower() == "description") or
          (key.lower() == "version") or
          (key.lower() == "user") or
          (key.lower() == "revision") or
          (key.lower() == "testenvironments") ):
         if( len(value) > 0):
            if( number_of_report_entries > 0 ):
               data += ",\n"
            data += "      " + quote + key + quote + ": " + quote + value + quote
            number_of_report_entries += 1
      elif( key.lower() == "start" ):
         if( len(value) > 0):
            if( number_of_report_entries > 0 ):
               data += ",\n"
            time = value.replace(" ", "T", 1)
            value = time + "+0000"
            data += "      " + quote + "startDate" + quote + ": " + quote + value + quote
            number_of_report_entries += 1
      elif( key.lower() == "finish" ):
         if( len(value) > 0):
            if( number_of_report_entries > 0 ):
               data += ",\n"
            time = value.replace(" ", "T", 1)
            value = time + "+0000"
            data += "      " + quote + "finishDate" + quote + ": " + quote + value + quote
            number_of_report_entries += 1
   i = 0
   data += "\n    },\n  " + quote + "tests" + quote + ": [{\n"
   while( i < number_of_catalog_entries ):
      key = catalog[i]
      value = catalog[i+1]
      i += 2
      if( (key.lower() == "testkey") or
          (key.lower() == "status") or
          (key.lower() == "comment") or
          (key.lower() == "examples") or
          (key.lower() == "defects") ):
         if( len(value) > 0):
            if( number_of_detail_entries > 0 ):
               data += ",\n"
            data += "      " + quote + key + quote + ": " + quote + value + quote
            number_of_detail_entries += 1
            number_of_report_entries += 1
      elif( (key.lower() == "start") or
            (key.lower() == "finish") ):
         if( len(value) > 0):
            if( number_of_detail_entries > 0 ):
               data += ",\n"
            time = value.replace(" ", "T", 1)
            value = time + "+0000"
            data += "      " + quote + key + quote + ": " + quote + value + quote
            number_of_detail_entries += 1
            number_of_report_entries += 1
      elif( key.lower() == "evidences" ):
         j = 0
         number_of_evidences = len( value )
         if( number_of_detail_entries > 0 ):
            data += ",\n"
         data += "      " + quote + key + quote + ": [{\n"
         filename = ""
         filetype = ""
         filedata = ""
         evidence_count = 0
         while( j < number_of_evidences ):
            evidence_key = value[j]
            evidence_value = value[j+1]
            if( evidence_key.lower() == "filename" ):
               filename = evidence_value
            elif( evidence_key.lower() == "contenttype" ):
               filetype = evidence_value
            elif( evidence_key.lower() == "data" ):
               filedata = evidence_value
            if( (len(filedata) > 0) and (len(filename) > 0) and (len(filetype) > 0) ):
               if( evidence_count > 0 ):
                  data += ",\n      {\n"
               data += "        " + quote + "filename" + quote + ": " + quote + filename + quote + ",\n"
               data += "        " + quote + "contentType" + quote + ": " + quote + filetype + quote + ",\n"
               data += "        " + quote + "data" + quote + ": " + quote + filedata + quote + "\n      }"
               evidence_count += 1
               filename = ""
               filetype = ""
               filedata = ""
            j += 2
         data += "\n      ]"
         number_of_detail_entries += 1
         number_of_report_entries += 1
      elif( key.lower() == "steps" ):
         j = 0
         number_of_steps = len( value )
         if( number_of_detail_entries > 0 ):
            data += ",\n"
         data += "      " + quote + key + quote + ": [{\n"
         step_status = ""
         step_text = ""
         step_count = 0
         while( j < number_of_steps ):
            step_key = value[j]
            step_value = value[j+1]
            if( step_key.lower() == "status" ):
               step_status = step_value
               step_text = ""
            elif( step_key.lower() == "comment" ):
               step_text = step_value.replace( '\\', '\\\\' ).replace( '"', r'\"' )
            if( (len(step_status) > 0) and (len(step_text) > 0) ):
               if( step_count > 0 ):
                  data += ",\n        {\n"
               data += "        " + quote + "status" + quote + ": " + quote + step_status + quote + ",\n      "
               data += quote + "comment" + quote + ": " + quote + step_text + quote + "\n        }"
               step_count += 1
               step_status = ""
               step_text = ""
            j += 2
         data += "\n      ]"
         number_of_detail_entries += 1
         number_of_report_entries += 1
   data += "\n    }]\n  }\n"
   if( number_of_report_entries > 0 ):
      pathname = datadir+"/"+GetFilename( "jira" )
      retval = retval | WriteFile( pathname, data )
      if( retval == 0 ):
         print( "==> Created [" + pathname + "] JSON file." )
         retval = SendMessage( pathname )
   else:
         print( "==> End Report Processing.  No entries to report." )
   return( retval )


def SendMessage( message ):
   retval = 0
   if ( send_to_jira ):
      tempfile = datadir+"/"+GetFilename( "temp" )
      jira_msg = "{\n  " + quote + "arguments" + quote + ":\n" + message + "}\n"
      jirafile = open( tempfile, "w" )
      jirafile.write( jira_msg )
      jirafile.close()
      command = cURLcmd + "@" + tempfile
      args = shlex.split( command )
      try:
         if( sys.version_info[0] < 3 ):
            proc = subprocess.Popen( command, shell=True,
                                     universal_newlines=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE )
         else:
            proc = subprocess.Popen( args, shell=True,
                                     universal_newlines=True,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE )
         (output, err) = proc.communicate()
         retval = proc.wait()
         if "< HTTP/1.1 400 Bad Request" in err:
            print( "==> ERROR: Bad data in cURL request." )
            retval = -1
         elif "couldn't connect to host" in err:
            print( "==> ERROR: Can't connect to host." )
            retval = -1
      except OSError:
         print( "==> ERROR: Invalid cURL command." )
         retval = -1
      if( retval == 0 ):
         print( "==> Sent Report to " + endpoint + "." )
      try:
         os.remove( tempfile )
      except Exception: 
         pass
   else:
      print( "==> Note: Jira messaging disabled." )
   if( (not cleanup_enabled) or (not send_to_jira) or (retval != 0) ):
      pathname = datadir+"/"+GetFilename( "jira" )
      rc = WriteFile( pathname, message )
      if( rc == 0 ):
         print( "==> Created [" + pathname + "] JSON file." )
   if( cleanup_enabled ):
      try:
         os.remove( manifestfile )
      except Exception: 
         pass
   return( retval )


def RunRules( rulesfile ):
   if( os.path.isfile(rulesfile) ):
      csvfile = os.path.basename( rulesfile )
      logfile   = logdir  + "/" + csvfile.rsplit(".csv", 1)[0] + ".log"
      if os.path.isfile(logfile):
         os.remove(logfile)
      ruleset = ReadRulesFile( rulesfile, logfile )
      if( ruleset ):
         return_value = ProcessRules( ruleset, logfile )
         localtime = getDateTime()
         if( return_value == 0 ):
            info = "==> PASS [" + rulesfile + "] - All tests succeeded at " + localtime 
            InfoLine( logfile, info )
         elif( return_value == -1 ):
            results = [-1]
            info = "==> EXIT [" + rulesfile + "] - Test execution stopped prior to completion at " + localtime
            InfoLine( logfile, info )
         else:
            info = "==> FAIL [" + rulesfile + "] - One or more test conditions failed at " + localtime
            InfoLine( logfile, info )
         results = ParseRulesLog( logfile )
      else:
         results = [-1]
         localtime = getDateTime()
         info = "==> ERROR: Rules file [" + rulesfile + "] is empty at " + localtime
   else:
      results = [-1]
      localtime = getDateTime()
      info = "==> ERROR: Rules file [" + rulesfile + "] does not exist at " + localtime
   return( results )


def ReadRulesLog( logfile ):
   lines = []
   try:
      log = open( logfile, "r" )
      for line in log:
         line = line.rstrip()
         lines.append( line )
      log.close
   except IOError:
      lines = [-1]
      print( "==> ERROR: [" + logfile + "] can't be opened for reading." )
   return( lines )


def ParseRulesLog( ruleslog ):
   result = 0
   results = []
   lines = ReadRulesLog( ruleslog )
   for line in lines:
      if( result ):
         test_result =     line[4:8]
         result_time =     line[-19:]
         sequence_number = result
         entry = "result: " + sequence_number + ",, " + test_result + ",, " + result_time
         results.append( entry.strip() )
      p = line.find( "Start run at" )
      if( p != -1 ):
         start_time =      line[-19:]
         entry = "start: " + start_time
         results.append( entry.strip() )
      p = line.find( "cmd:" )
      if( p != -1 ):
         sequence_number = line[:3]
         command =         line[9:]
         entry = "command: " + sequence_number + ",, " + command
         results.append( entry.strip() )
      p = line.find( "tst:" )
      if( p != -1 ):
         sequence_number = line[:3]
         test =            line[9:]
         result = sequence_number
         entry = "test: " + sequence_number + ",, " + test
         results.append( entry.strip() )
      else:
         result = 0
      p = line.find( "if contains" )
      if( p != -1 ):
         container = line.split(" ")
         exhibit = "exhibit: " + str(container[-1])
         results.append( exhibit.strip() )
   exhibit = "exhibit: " + ruleslog
   results.append( exhibit.strip() )
   end = "end: " + line[4:8] + ",, " + line[-19:]
   results.append( end.strip() )
   return( results )

