#!/usr/bin/python

'''
   File name: TlogMetr.py
   Author: Wayne Parham
   Date created: 09/01/2019
'''


from __future__ import division
from TlogLib import *
import string, sys


def TransactionCount( tloglist, date=None ):
   count = 0
   for line in tloglist:
      for record in line:
         try:
            stringtype = hex(ord(record[0]))
            if( stringtype == "0x0" ):
               transdate = BCD2string( record[11] )
               if( date ):
                  if( transdate == date ):
                     count += 1
               else:
                  count += 1
         except:
            pass
   return( count )


def Print00details( asciir ):
   print( "Date \t\t" +         asciir[11] )
   print( "Time \t\t" +         asciir[12] )
   print( "Transaction \t" +    asciir[3]  )
   print( "Type \t\t" +         asciir[4]  )
   print( "Record count \t" +   asciir[5]  )
   print( "Terminal \t" +       asciir[1]  )
   print( "Operator \t" +       asciir[2]  )
   print( "Password \t" +       asciir[10] )
   print( "Gross Positive \t" + asciir[6]  )
   print( "Gross Negative \t" + asciir[7]  )
   print( "Net cash \t" +       asciir[8]  )
   print( "Non cash \t" +       asciir[9]  )
   print( "Ring Time \t" +      asciir[15] )
   print( "Tender Time \t" +    asciir[16] )
   print( "Other Time \t" +     asciir[17] )
   print( "Inactive_Time \t" +  asciir[22] )
   print( "Previous Time \t" +  asciir[18] )
   print( "Tenant Number \t" +  asciir[20] )
   print( "Indicator 0 \t" +    asciir[13] )
   print( "Indicator 1 \t" +    asciir[14] )
   print( "Indicator 2 \t" +    asciir[19] )
   print( "Token \t\t" +        asciir[23] )
   print( "Token Type \t" +     asciir[24] )
   return


def TransactionDetails( tloglist, date=None, transaction=None ):
   count = 0
   field_types = []
   for line in tloglist:
      for record in line:
         try:
            stringtype, subtype = Identify( record )
            if( stringtype == "0x0" ):
               count += 1
               asciir = Asciify( record )
               transaction_number = asciir[3]
               transaction_date   = asciir[11]
               if( date and transaction ):
                  if( (date == transaction_date) and (transaction == transaction_number) ):
                     print( "\n" )
                     Print00details( asciir )
               elif( transaction ):
                  if( transaction == transaction_number ):
                     print( "\n" )
                     Print00details( asciir )
               elif( date ):
                  if( date == transaction_date ):
                     Print00details( asciir )
                     print( "\n=====================================\n" )
               else:
                  Print00details( asciir )
                  print( "\n=====================================\n" )
         except:
            pass
   return( count )


def ItemCount( tloglist, date=None, transaction=None ):
   cumulative_tender_time = 0
   cumulative_other_time = 0
   cumulative_ring_time = 0
   transaction_count = 0
   item_count = 0
   field_types = []
   for line in tloglist:
      transaction_record = False
      for record in line:
         try:
            stringtype, subtype = Identify( record )
            if( stringtype == "0x0" ):
               asciir = Asciify( record )
               transaction_number = asciir[3]
               transaction_date   = asciir[11]
               ring_time          = asciir[15]
               tender_time        = asciir[16]
               other_time         = asciir[17]
               if( date and transaction ):
                  if( (date == transaction_date) and (transaction == transaction_number) ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # print( "\n" )
                     # Print00details( asciir )
               elif( transaction ):
                  if( transaction == transaction_number ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # print( "\n" )
                     # Print00details( asciir )
               elif( date ):
                  if( date == transaction_date ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # Print00details( asciir )
                     # print( "\n=====================================\n" )
               else:
                  transaction_record = True
                  transaction_count += 1
                  cumulative_ring_time += int(ring_time)
                  cumulative_tender_time += int(tender_time)
                  cumulative_other_time += int(other_time)
                  # Print00details( asciir )
                  # print( "\n=====================================\n" )
            if( transaction_record and stringtype == "0x6" ):
               if( transaction ):
                  if( transaction == transaction_number ):
                     item_count += 1
               else:
                  item_count += 1
         except:
            pass
   return( transaction_count, item_count, cumulative_ring_time,
           cumulative_tender_time, cumulative_other_time )


def CarePlanCount( tloglist, date=None, transaction=None, details=None ):
   cumulative_wpp_weighted_time = 0
   cumulative_tender_time = 0
   cumulative_other_time = 0
   cumulative_ring_time = 0
   cumulative_item_count = 0
   cumulative_plan_count = 0
   transaction_count = 0
   max_wpp_time = 0
   item_count = 0
   plan_count = 0
   field_types = []
   if( details):
      print( "# date,\t time,\t trans,\t items,\t plans,\t ringtime,\t wpptime" )
   for line in tloglist:
      transaction_record = False
      contains_WPP = False
      for record in line:
         try:
            stringtype, subtype = Identify( record )
            if( stringtype == "0x0" ):
               item_count = 0
               plan_count = 0
               wpp_weighted_time = 0
               asciir = Asciify( record )
               transaction_number = asciir[3]
               transaction_date   = asciir[11]
               transaction_time   = asciir[12]
               ring_time          = asciir[15]
               tender_time        = asciir[16]
               other_time         = asciir[17]
               if( date and transaction ):
                  if( (date == transaction_date) and (transaction == transaction_number) ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # print( "\n" )
                     # Print00details( asciir )
               elif( transaction ):
                  if( transaction == transaction_number ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # print( "\n" )
                     # Print00details( asciir )
               elif( date ):
                  if( date == transaction_date ):
                     transaction_record = True
                     transaction_count += 1
                     cumulative_ring_time += int(ring_time)
                     cumulative_tender_time += int(tender_time)
                     cumulative_other_time += int(other_time)
                     # Print00details( asciir )
                     # print( "\n=====================================\n" )
               else:
                  transaction_record = True
                  transaction_count += 1
                  cumulative_ring_time += int(ring_time)
                  cumulative_tender_time += int(tender_time)
                  cumulative_other_time += int(other_time)
                  # Print00details( asciir )
                  # print( "\n=====================================\n" )
            if( transaction_record and stringtype == "0x6" ):
               if( transaction ):
                  if( transaction == transaction_number ):
                     item_count += 1
                     cumulative_item_count += 1
               else:
                  item_count += 1
                  cumulative_item_count += 1
            if( transaction_record and stringtype == "0x15" and subtype == "0x13" ):
               wpp_upc = record[2]
               company_code = record[5]
               if( company_code == b'13' ):
                  # print( record )
                  contains_WPP = True
                  if( transaction ):
                     if( transaction == transaction_number ):
                        plan_count += 1
                        cumulative_plan_count += 1
                  else:
                     plan_count += 1
                     cumulative_plan_count += 1
         except:
            pass
      if( contains_WPP ):
         wpp_ratio = plan_count / item_count
         wpp_weighted_time = wpp_ratio * int(ring_time)
         cumulative_wpp_weighted_time += wpp_weighted_time
         if( wpp_weighted_time > max_wpp_time ):
            max_wpp_time = wpp_weighted_time
         if( details):
            print( transaction_date + ",\t" + transaction_time + ",\t" + transaction_number + ",\t" +
                   str(item_count) + ",\t" + str(plan_count) + ",\t" + str(int(ring_time)) + ",\t" +
                   str(round(wpp_weighted_time,2)) )
   return( transaction_count, cumulative_item_count, cumulative_plan_count, cumulative_ring_time,
           cumulative_tender_time, cumulative_other_time, cumulative_wpp_weighted_time, max_wpp_time )

